local SpineHelper = require('spiner-helper')

local module = {states={}}

function module.load(x, y)
    local self = {
        name = 'creature',
        body = love.physics.newBody(world, x, y, 'dynamic'),
        aura = {
            shape = love.physics.newCircleShape(300),
        },
        attack = {
            shape = love.physics.newCircleShape(30),
        },
    }
    self.body:setMass(1)
    self.aura.fixture = love.physics.newFixture(self.body, self.aura.shape)
    self.aura.fixture:setSensor(true)
    self.aura.fixture:setUserData({
        name = 'creature-aura',
        obj = self
    })
    self.aura.fixture:setGroupIndex(-1)
    self.attack.fixture = love.physics.newFixture(self.body, self.attack.shape)
    self.attack.fixture:setSensor(true)
    self.attack.fixture:setUserData({
        name = 'creature-attack',
        obj = self
    })
    self.attack.fixture:setGroupIndex(-1)

    self.spine = SpineHelper.load('panther/panther.json', 0.25,
        function(self, attachment)
            return love.graphics.newImage('panther/'.. attachment.name .. '.png')
        end)
    self.spine.skeleton.debugSlots = false


    return setmetatable(self, {
        __index = function(obj, key)
            return module[key]
        end
    })
end

function module:update(dt)
    -- local mx, my = love.mouse.getPosition()
    -- self.body:setPosition(
    --     (mx - WIDTH/(2*zoom) + player.skeleton.x) / zoom,
    --     (my - HEIGHT/(2*zoom) + player.skeleton.y - 134) / zoom
    -- )
    self.state:update(self, dt)
    local x, y = self.body:getPosition()
    if self.spine.skeleton.x ~= x then
        self.spine.skeleton.flipX = x < self.spine.skeleton.x
    end
    self.spine.skeleton.x, self.spine.skeleton.y = x, y
    self.spine:update(dt)
end

function module:transitionTo(stateName, ...)
    local state = self.states[stateName]
    if self.state ~= state then
        print(string.format('Transitioning from %s to %s',
            ((self.state or {}).name or 'nil'), ((state or {}).name or 'nil')))
        state:enter(self, ...)
        self.state = state
    end
end

module.states.approach = {name = 'approach'}
function module.states.approach:enter(creature, dir)
    creature.spine:setAnimation('walking', true)
    if dir ~= nil then
        creature.dir = dir
    end
    creature.body:setLinearVelocity(creature.dir * 80, 0)
    self.jumpTimer = 0
    self.jumpDelay = 3
end
function module.states.approach:update(creature, dt)
    self.jumpTimer = self.jumpTimer + dt
    if self.jumpTimer > self.jumpDelay then
        if math.random() < 0.7 then
            creature:transitionTo('pause')
        else
            creature:transitionTo('jump')
        end
    elseif creature.body:getX() < 100 then
        creature.dir = 1
        creature.body:setLinearVelocity(creature.dir * 80, 0)
    end
end

local function jumpCoroutine(obj, height)
    local initalSpeed = -math.sqrt(2 * GRAVITY * height)
    local duration = -2 * initalSpeed / GRAVITY
    obj.body:setLinearVelocity(-obj.dir * initalSpeed, initalSpeed)
    return coroutine.create(function()
        local dt
        local vy
        local timer = 0
        while timer < duration do
            dt = coroutine.yield()
            timer = timer + dt
            vy = GRAVITY * timer + initalSpeed
            obj.body:setLinearVelocity(-obj.dir * initalSpeed, vy)
        end
    end)
end


module.states.jump = {name = 'jump'}
function module.states.jump:enter(creature)
    local height = math.random(20, 100)
    print('jumping', height)
    self.co = jumpCoroutine(creature, height)
end
function module.states.jump:update(creature, dt)
    if coroutine.status(self.co) ~= 'dead' then
        coroutine.resume(self.co, dt)
    else
        creature:transitionTo('approach')
    end
end

module.states.pause = {name = 'pause'}
function module.states.pause:enter(creature)
    creature.body:setLinearVelocity(0, 0)
    creature.spine:setAnimation('idle', true)
    self.duration = math.random(1, 3)
    self.timer = 0
end
function module.states.pause:update(creature, dt)
    self.timer = self.timer + dt
    if self.timer > self.duration then
        creature:transitionTo('approach')
    end
end

function module:draw()
    self.spine:draw()
    local x, y = self.body:getPosition()
    love.graphics.setColor(255, 0, 0)
    love.graphics.circle('line', x, y, self.attack.shape:getRadius())
    love.graphics.setColor(255, 255, 0)
    love.graphics.circle('line', x, y, self.aura.shape:getRadius())
end

return module
